##
## Voices for NVDA
## Copyright (c) 2015 by Code Factory, S.L. All rights reserved.
## Eloquence and Vocalizer Expressive are registered trademarks
## of Nuance Communications, Inc.
##

import os
import os.path
from ctypes import *
from ctypes.wintypes import *
import time
import threading
import Queue
from collections import OrderedDict

import nvwave
import config
from logHandler import log

import addonHandler


addonHandler.initTranslation()


## Mapping NVDAInterface.h to Python
class WAVEFORMATEX(Structure):
    _fields_ = [
        ("wFormatTag", WORD),
        ("nChannels", WORD),
        ("nSamplesPerSec", DWORD),
        ("nAvgBytesPerSec", DWORD),
        ("nBlockAlign", WORD),
        ("wBitsPerSample", WORD),
        ("cbSize", WORD)
    ]  
LPWAVEFORMATEX = POINTER(WAVEFORMATEX)

## The required SAPI Events and structures
SPEI_UNDEFINED = 0
SPEI_START_INPUT_STREAM = 1
SPEI_END_INPUT_STREAM = 2
SPEI_VOICE_CHANGE = 3
SPEI_TTS_BOOKMARK = 4
SPEI_WORD_BOUNDARY = 5
SPEI_PHONEME = 6
SPEI_SENTENCE_BOUNDARY = 7
SPEI_VISEME = 8
SPEI_TTS_AUDIO_LEVEL = 9

SPEVENTNUM = [
    "SPEI_UNDEFINED",
    #--- TTS engine
    "SPEI_START_INPUT_STREAM",
    "SPEI_END_INPUT_STREAM",
    "SPEI_VOICE_CHANGE",
    "SPEI_TTS_BOOKMARK",
    "SPEI_WORD_BOUNDARY",
    "SPEI_PHONEME",
    "SPEI_SENTENCE_BOUNDARY",
    "SPEI_VISEME",
    "SPEI_TTS_AUDIO_LEVEL",  
]  

class SPEVENT(Structure):
    _fields_ = [    
        ("eEventId", WORD ),
        ("elParamType", WORD ),
        ("ulStreamNum", ULONG ),
        ("ullAudioStreamOffset", ULARGE_INTEGER ),
        ("wParam", WPARAM ),
        ("lParam", LPARAM )
    ]
    
MAX_NVDA_STRING = 128

NVDA_VOICE_QUALITY = [ "Compact",        # 0
                       "Standard",      # 1
                       "Plus",      # 2
                       "Premium",        # 3
                       "PremiumHigh" ]    # 4

class NVDA_VOICE_INFO(Structure):
    _fields_ = [
        ("Language", WCHAR * MAX_NVDA_STRING ),
        ("Country", WCHAR * MAX_NVDA_STRING ),
        ("Variant", WCHAR * MAX_NVDA_STRING ),
        ("VoiceName", WCHAR * MAX_NVDA_STRING ),
        ("Quality",c_int),
        ("SAPILanguage",DWORD)
    ]
LPNVDA_VOICE_INFO = POINTER(NVDA_VOICE_INFO)

class NVDA_EVENT(Structure):
    _fields_ = [
        ("SAPIEvent", SPEVENT ),
        ("AudioData", POINTER(BYTE)),
        ("AudioDataLen", UINT)
    ]    
LPNVDA_EVENT = POINTER(NVDA_EVENT)

## Voice Profiles
NVDAProfile_Reed = 0
NVDAProfile_Shelly = 1
NVDAProfile_Bobby = 2
NVDAProfile_Rocko = 3
NVDAProfile_Glen = 4
NVDAProfile_Sandy = 6
NVDAProfile_Grandma = 6
NVDAProfile_Grandpa = 7

NVDA_ETI_VOICEPROFILES = [
    "Reed",
    "Shelly", 
    "Bobby", 
    "Rocko", 
    "Glen", 
    "Sandy", 
    "Grandma", 
    "Grandpa"
]


#### Globals...



lastIndex = None

_isSpeaking = False
_bgThread = None
_bgQueue = None
_player = None
_ttsDLL = None
_initQ = None
_voiceHash = None
_voiceList = None
_variantHash = None
_variantList = None


def _log_enter(fn,msg):
   now = time.time()
   localtime = time.localtime(now)
   milliseconds = '%03d' % int((now - int(now)) * 1000)
   prefix =  time.strftime('%Y-%m-%d:%H:%M:%S.', localtime) + milliseconds
   log.info("CF: %s: ENTER NVDA%s : %s" % ( prefix, fn, msg ))

def _log_exit(fn, ret):
   now = time.time()
   localtime = time.localtime(now)
   milliseconds = '%03d' % int((now - int(now)) * 1000)
   prefix =  time.strftime('%Y-%m-%d-%H:%M:%S.', localtime) + milliseconds
   log.info("CF: %s: EXIT  NVDA%s %s" % ( prefix, fn, str(ret)))



class  CodeFactoryException(Exception):
    pass

### A seperate thread for TTS + nvwave
## The callback handler, playback and SAPI events.
## Plays directly to nvwave
codefactory_callback=CFUNCTYPE(c_int,LPNVDA_EVENT)

@codefactory_callback
def cfcallback(nvda_event):
    global _player, _isSpeaking, lastIndex
    try:
        if not _isSpeaking:
            return 1
        numsamples = nvda_event.contents.AudioDataLen
        wav = nvda_event.contents.AudioData
        if numsamples > 0:
            try:
                ##_log_enter("player::feed",str(numsamples))
                _player.feed(string_at(wav, numsamples))
                ##_log_exit("player::feed",0)
            except:
                log.debugWarning("Error feeding audio to nvWave",exc_info=True)
        e  = nvda_event.contents.SAPIEvent.eEventId
        if e == SPEI_TTS_BOOKMARK:
            x = wstring_at(nvda_event.contents.SAPIEvent.lParam)
            lastIndex = int(x)
    except:
        log.error("callback", exc_info=True)
    return 0

class BgThread(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
        self.setDaemon(True)
        
    def run(self):
        global _bgQueue
        while True:
            func, args, kwargs = _bgQueue.get()
            if not func:
                break
            try:
                func(*args, **kwargs)
            except:
                log.error("Error running function from queue", exc_info=True)
                _bgQueue.task_done()

def _execWhenDone(func, *args, **kwargs):
    global _bgQueue
    ##_log_enter("ENQUEUE",str(func)+str(args))        
    mustBeAsync = kwargs.pop("mustBeAsync", False)
    if mustBeAsync or _bgQueue.unfinished_tasks != 0:
        _bgQueue.put((func, args, kwargs))
    else:
        func(*args, **kwargs)
    ##_log_exit("ENQUEUE",_bgQueue.qsize())

def _speak(text):
	global _isSpeaking, _ttsDLL,_player
	_isSpeaking = True
        ##_log_enter("Speak",text)
        ret = _ttsDLL.NVDASpeak(text, cfcallback)
        ##_log_exit("Speak",ret)
        _isSpeaking = False
        _player.idle()
        
def speak(text):
    global _bgQueue
    _execWhenDone( _speak, text, mustBeAsync = True)

def stop():
    global _isSpeaking, _bgQueue, lastIndex, _player
    others =  []
    _isSpeaking = False
    _player.stop()
    try:
        while True:
            item = _bgQueue.get_nowait()
            if item[0] != _speak:
                others.append(item)
            _bgQueue.task_done()
    except Queue.Empty:
        pass
    for item in others:
        _bgQueue.put(item)
    lastIndex=None

## pauses and continues.    
def pause(switch):    
    global _player
    _player.pause(switch)




    
## Initialize
def initialize(name, dll,exe):
    global _ttsDLL, _bgQueue, _bgThread, _player, _initQ
    ## try to use embedded or system wide installation
    if not os.path.exists(dll):
        raise CodeFactoryException(" Can't find DLL : %s" % dll)
    ## Rest of initialize in thread...
    _bgQueue = Queue.Queue()
    _bgThread=BgThread()
    _bgThread.start()
    #_initQ = Queue.Queue()
    #_execWhenDone(_initalize,dll,mustBeAsync=True)
    ret = _initalize(dll,name,exe)
    #ret = _initQ.get()
    #_initQ.task_done
    if ( ret != 0 ):
        raise CodeFactoryException("Initializing %s Filed" % name)

def _initalize(dllpath,name,exename):
    ## Load with ctypes, and initialize funtions
    global _ttsDLL, _player, _initQ
    try:
        ##_log_enter("LoadLibrary",dllpath)
        _ttsDLL = cdll.LoadLibrary(dllpath)
        ##_log_exit("LoadLibrary",_ttsDLL)
        ## Setup the ctypes definitions .. can be removed.
        _ttsDLL.NVDAInitialize.argtypes = ()
        _ttsDLL.NVDAInitialize.restypes = c_int
        ##
        _ttsDLL.NVDAGetAudioFormat.argtypes = [ LPWAVEFORMATEX ]
        _ttsDLL.NVDAGetAudioFormat.restypes = c_int
        ##
        _ttsDLL.NVDAGetVoiceList.argtypes = [ LPNVDA_VOICE_INFO, UINT ]
        _ttsDLL.NVDAGetVoiceList.restypes = UINT
        ##
        _ttsDLL.NVDASetCurrentVoice.argtypes = [ LPNVDA_VOICE_INFO ]
        _ttsDLL.NVDASetCurrentVoice.restypes = UINT
        #_ttsDLL.NVDASpeak
        _ttsDLL.NVDARelease.argtypes = []
        _ttsDLL.NVDARelease.restypes = c_int
        _ttsDLL.NVDASetMasterRate.argtypes = [ c_int ]
        _ttsDLL.NVDASetMasterRate.restypes = c_int 
        _ttsDLL.NVDASetMasterPitch.argtypes = [ c_int ]
        _ttsDLL.NVDASetMasterPitch.restypes = c_int 
        _ttsDLL.NVDASetMasterVolume.argtypes = [ c_int ]
        _ttsDLL.NVDASetMasterVolume.restypes = c_int 
        _ttsDLL.NVDASetVoiceProfile.argtypes = [ c_int ]
        _ttsDLL.NVDASetVoiceProfile.rettypes = c_int
        _ttsDLL.NVDAStop.argtypes = []
        _ttsDLL.NVDAStop.restypes = c_int
        _ttsDLL.NVDASetAudioBufferSize.argtypes = [ UINT ]
        _ttsDLL.NVDASetAudioBufferSize.restypes = [ c_int ]
        _ttsDLL.NVDAIsAuthorized.argtypes = []
        _ttsDLL.NVDAIsAuthorized.restypes = c_int
        #_ttsDLL.NVDALaunchEngineUI.argtypes = []
        #_ttsDLL.NVDALaunchEngineUI.restypes = c_int
        ## Now do the init...
        ##_log_enter("Initialize","")
        init_code = _ttsDLL.NVDAInitialize()
        init_code = _ttsDLL.NVDAIsAuthorized()
        if init_code != 1 :
            import wx
            import gui
            def msg(exename):
                gui.messageBox(
                    _("You need to register your trial or permanent license before being able to use Code Factory TTS.\nPressing OK will start the registration software.") ,
                    "Voices for NVDA",wx.OK)
                os.startfile(exename)
            wx.CallAfter(msg,exename)
            log.info("%s\nUnregistered copy of %s; Invoke UI from NVDA Context Menu\n%s" % ( "*"*50,name,"*"*50 ))
            raise CodeFactoryException("Unregistered copy")

        init_code = _initializeVoiceData()
        #_initQ.put(0)
        if not init_code:
            import wx
            import gui
            def msg(exename):
                gui.messageBox(
                    _("You need to install %s voices to use Code Factory TTS.\nPressing OK will start voice installation software.") % name,
                    "Voices for NVDA",wx.OK)
                os.startfile(exename)
            wx.CallAfter(msg,exename)
            log.info("%s\nNo voices installed in %s; Invoke UI from NVDA Context Menu\n%s" % ('*'*50,name,'*'*50 ))
            raise CodeFactoryException("No voices installed")
        
        ##_log_exit("Initialize",init_code)
        fmt = WAVEFORMATEX()
        ret_code = _ttsDLL.NVDAGetAudioFormat(LPWAVEFORMATEX(fmt))
        ## Make buffer for 300ms
        bufsize = ( int(fmt.nAvgBytesPerSec * 0.300 / 1024) +1 ) * 1024
        _ttsDLL.NVDASetAudioBufferSize(bufsize)
        
        _player = nvwave.WavePlayer(channels=fmt.nChannels,
                                    samplesPerSec=fmt.nSamplesPerSec,
                                    bitsPerSample=fmt.wBitsPerSample,
                                outputDevice=config.conf["speech"]["outputDevice"])
        return 0
    except:
        del _ttsDLL
	log.error("TTS loading", exc_info=True)
        #_initQ.put(1)        
        raise CodeFactoryException( dllpath + " is not valid")

def terminate():
    global _ttsDLL, _voiceList, _voiceHash, _variantList, _variantHash
    global _bgQueue, _bgThread, _player
    _ttsDLL.NVDAStop()
    stop()
    _bgQueue.put((None, None, None))
    _bgThread.join()
    ##_log_enter("Release","")
    _ttsDLL.NVDARelease()
    ##_log_exit("Release",0)
    _player.close()
    del _ttsDLL
    
    _ttsDLL = None
    _voiceHash = None
    _voiceList = None
    _bgThread = None
    _bgQueue = None
    _variantHash = None
    _variantList = None
    
def _generateVoiceName(l,c,va,vo,q):
    return "%s %s [%s] (%s)" % ( vo, va, l, NVDA_VOICE_QUALITY[q])

def _initializeVoiceData():
    ## Prepare Voice Hash.
    global _ttsDLL, _voiceHash, _voiceList, _variantList, _variantHash
    voices_size = _ttsDLL.NVDAGetVoiceList(None,0)    
    voices_count = voices_size / sizeof(NVDA_VOICE_INFO)
    if voices_count == 0:
        return False
    voices = (NVDA_VOICE_INFO*voices_count)()
    _voiceHash = OrderedDict()
    _voiceList = []
    _variantHash = OrderedDict()
    _variantList = []
    index = 0
    ret_code = _ttsDLL.NVDAGetVoiceList(LPNVDA_VOICE_INFO(voices),
                                       sizeof(voices))
    
    ## Language/Country/Variant/VoiceName/NVDA_VOICE_INFO[Quality]
    for v in voices:
        ## Language/Country/Variant/VoiceName/NVDA_VOICE_INFO[Quality]
        ## VoiceName Variant [Language] (Quality)
        id = _generateVoiceName(v.Language, 
                                v.Country,
                                v.Variant,
                                v.VoiceName,
                                v.Quality )
        _voiceHash[id] = ( index ,
                           v.Language, 
                           v.Country,
                           v.Variant,
                           v.VoiceName,
                           v.Quality,
                           v.SAPILanguage )
        _voiceList.append( (id,v.VoiceName,v.Language)  )
        index = index + 1
    ## TODO: do this for only Eloq ? will this harm Voca?
    for v in range(0,len(NVDA_ETI_VOICEPROFILES)):
        id = NVDA_ETI_VOICEPROFILES[v]
        _variantHash[id] = ( v, id )
        _variantList.append( (id,id))
    ##
    return True

## Return [ (ID / name / language ) ]
def getVoiceList():
    global _voiceList
    return _voiceList
## Expect the string created above...
def _setVoice(idx):
    global _voiceHash
    tpl = _voiceHash[idx]
    vi = NVDA_VOICE_INFO()
    vi.Language = tpl[1]
    vi.Country = tpl[2]
    vi.Variant = tpl[3]
    vi.VoiceName = tpl[4]
    vi.Quality = tpl[5]
    vi.SAPILanguage = tpl[6]
    ret = _ttsDLL.NVDASetCurrentVoice( LPNVDA_VOICE_INFO(vi))

def setVoice(idx):
    global _voiceHash
    if idx not in _voiceHash:
        raise CodeFactoryException("voice not found. Using default")
    else:
        _execWhenDone( _setVoice, idx)    
## Return [ (ID / name ) ]
def getVariantList():
    global _variantList
    return _variantList

def _setVariant(idx):
    global _variantHash, _ttsDLL
    var = _variantHash[idx][0]
    _ttsDLL.NVDASetVoiceProfile( var)

def setVariant(idx):    
    _execWhenDone( _setVariant, idx)

def _setPitch(p):
    global _ttsDLL
    _ttsDLL.NVDASetMasterPitch(p)
    
def setPitch(p):    
    _execWhenDone( _setPitch, p)

def _setRate(r):
    global _ttsDLL
    _ttsDLL.NVDASetMasterRate(r)
    
def setRate(r):    
    _execWhenDone( _setRate, r)
    
def _setVolume(v):
    global _ttsDLL
    _ttsDLL.NVDASetMasterVolume(v)
    
def setVolume(v):
    _execWhenDone( _setVolume, v)    
